"""GitHub Action wrapper."""

from __future__ import annotations

import logging
import typing

import attr

# pylint: disable=relative-beyond-top-level
from ..api.app_client import GitHubApp
# pylint: disable=relative-beyond-top-level
from ..api.raw_client import RawGitHubAPI
# pylint: disable=relative-beyond-top-level
from ..api.tokens import GitHubOAuthToken
# pylint: disable=relative-beyond-top-level,import-error
from ..models.events import GidgetHubActionEvent


if typing.TYPE_CHECKING:
    # pylint: disable=relative-beyond-top-level
    from ...app.action.config import GitHubActionConfig


logger = logging.getLogger(__name__)


@attr.dataclass
class GitHubAction(GitHubApp):
    """GitHub Action API wrapper."""

    _metadata: GitHubActionConfig = attr.ib(default=None)
    """A GitHub Action metadata from envronment vars."""

    @_metadata.validator
    def _verify_metadata_is_set(self, attribute, value):
        if value is None:
            raise ValueError(f'{attribute} must be set.')

    @property
    def event(self):  # noqa: D401
        """Parsed GitHub Action event data."""
        return GidgetHubActionEvent.from_file(
            self._metadata.event_name,  # pylint: disable=no-member
            self._metadata.event_path,  # pylint: disable=no-member
        )

    @property
    def token(self):
        """Return GitHub Action access token."""
        return GitHubOAuthToken(
            self._metadata.token,  # pylint: disable=no-member
        )

    @property
    def api_client(self):  # noqa: D401
        """The GitHub App client."""
        return RawGitHubAPI(
            token=self.token,
            session=self._http_session,
            user_agent=self._config.user_agent,
        )



"""Tests for GitHub private key class."""
import random
import re
from datetime import date
from pathlib import Path

import pytest

from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from jwt import decode as parse_jwt

from octomachinery.github.models.private_key import GitHubPrivateKey


@pytest.fixture
def rsa_public_key(rsa_private_key):
    """Extract a public key out of private one."""
    return rsa_private_key.public_key()


@pytest.fixture
def rsa_public_key_bytes(rsa_public_key) -> bytes:
    """Return a PKCS#1 formatted RSA public key encoded as PEM."""
    return rsa_public_key.public_bytes(
        encoding=Encoding.PEM,
        format=PublicFormat.PKCS1,
    )


@pytest.fixture
def github_private_key(rsa_private_key_bytes: bytes) -> GitHubPrivateKey:
    """Construct a test instance of ``GitHubPrivateKey``."""
    return GitHubPrivateKey(rsa_private_key_bytes)


@pytest.fixture
def rsa_private_key_path(
        rsa_private_key_bytes: bytes,
        tmp_path_factory,
) -> Path:
    """Save the private key to disk as PEM file and provide its path."""
    tmp_dir = tmp_path_factory.mktemp('github_private_key')
    private_key_filename = (
        f'test-github-app.{date.today():%Y-%m-%d}'
        '.private-key.pem'
    )
    private_key_path = tmp_dir / private_key_filename
    private_key_path.write_bytes(rsa_private_key_bytes)
    return private_key_path


def test_github_private_key__from_file(
        github_private_key,
        rsa_private_key_path: Path,
):
    """Test that GitHubPrivateKey from file and bytes are the same."""
    key_from_file = GitHubPrivateKey.from_file(rsa_private_key_path)
    assert key_from_file == github_private_key


def test_github_private_key____repr__(github_private_key):
    """Verify what repr protocol only exposes fingerprint."""
    repr_pattern = re.compile(
        r"^<GitHubPrivateKey\(b_raw_data=b'<SECRET>'\)\s"
        r'with\sSHA\-1\sfingerprint\s'
        r"'[a-f0-9]{2}(:[a-f0-9]{2}){19}'>$",
    )
    assert repr_pattern.match(repr(github_private_key))
    assert repr_pattern.match(f'{github_private_key!r}')


def test_github_private_key____str__(github_private_key):
    """Verify that the string protocol doesn't expose secrets."""
    escaped_private_key_repr = (
        repr(github_private_key).
        replace('(', r'\(').
        replace(')', r'\)')
    )
    exception_message = (
        "<class 'octomachinery.github.models.private_key.GitHubPrivateKey'> "
        'objects do not implement the string protocol '
        'for security reasons. '
        f'The repr of this instance is {escaped_private_key_repr!s}.'
    )
    with pytest.raises(TypeError, match=exception_message):
        str(github_private_key)
    with pytest.raises(TypeError, match=exception_message):
        f'{github_private_key!s}'  # pylint: disable=pointless-statement


def test_github_private_key__make_jwt_for(
        github_private_key: GitHubPrivateKey,
        rsa_public_key_bytes,
):
    """Verify that e2e encoding-decoding of the JWT works."""
    github_app_id = random.randint(0, 9999999)
    jwt_string = github_private_key.make_jwt_for(app_id=github_app_id)
    payload = parse_jwt(
        jwt_string.encode('utf-8'), rsa_public_key_bytes, algorithms='RS256',
    )
    assert payload['iss'] == github_app_id
    assert payload['exp'] - payload['iat'] == 60


def test_github_private_key__make_jwt_for__invalid_timeout(github_private_key):
    """Verify that time offset can't exceed 10 mins."""
    github_app_id = random.randint(0, 9999999)
    with pytest.raises(
            ValueError,
            match='The time offset must be less than 10 minutes',
    ):
        github_private_key.make_jwt_for(app_id=github_app_id, time_offset=601)



"""Compatibility shims for the models subpackage."""
from functools import wraps as _wraps_function

from jwt import encode as _compute_jwt


try:
    from jwt import __version__ as _pyjwt_version_str
except ImportError:
    _pyjwt_version_str = '0.0.0'


_pyjwt_version_info = tuple(map(int, _pyjwt_version_str.split('.')))
_is_pyjwt_above_v2_0 = _pyjwt_version_info >= (2, 0, 0)


@_wraps_function(_compute_jwt)
def _compute_jwt_below_v2_0(*args, **kwargs) -> str:
    return _compute_jwt(*args, **kwargs).decode('utf-8')


compute_jwt = _compute_jwt if _is_pyjwt_above_v2_0 else _compute_jwt_below_v2_0



"""Tests for circular imports in all local packages and modules.

This ensures all internal packages can be imported right away without
any need to import some other module before doing so.

This module is based on an idea that pytest uses for self-testing:
* https://github.com/pytest-dev/pytest/blob/d18c75b/testing/test_meta.py
* https://twitter.com/codewithanthony/status/1229445110510735361
"""
import os
import pkgutil
import subprocess
import sys
from itertools import chain
from pathlib import Path

import pytest

import octomachinery


def _find_all_importables(pkg):
    """Find all importables in the project.

    Return them in order.
    """
    return sorted(
        set(
            chain.from_iterable(
                _discover_path_importables(Path(p), pkg.__name__)
                for p in pkg.__path__
            ),
        ),
    )


def _discover_path_importables(pkg_pth, pkg_name):
    """Yield all importables under a given path and package."""
    for dir_path, _d, file_names in os.walk(pkg_pth):
        pkg_dir_path = Path(dir_path)

        if pkg_dir_path.parts[-1] == '__pycache__':
            continue

        if all(Path(_).suffix != '.py' for _ in file_names):
            continue

        rel_pt = pkg_dir_path.relative_to(pkg_pth)
        pkg_pref = '.'.join((pkg_name,) + rel_pt.parts)
        yield from (
            pkg_path
            for _, pkg_path, _ in pkgutil.walk_packages(
                (str(pkg_dir_path),), prefix=f'{pkg_pref}.',
            )
        )


@pytest.mark.parametrize(
    'import_path',
    _find_all_importables(octomachinery),
)
def test_no_warnings(import_path):
    """Verify that exploding importables doesn't explode.

    This is seeking for any import errors including ones caused
    by circular imports.
    """
    imp_cmd = (
        sys.executable,
        '-W', 'error',

        # NOTE: These are necessary for `tox -e old-deps`:
        '-W', "ignore:Using or importing the ABCs from 'collections' instead "
        "of from 'collections.abc' is deprecated since "
        'Python 3.3, and in 3.10 it will stop working:'
        'DeprecationWarning:jwt.api_jwt',
        '-W', "ignore:Using or importing the ABCs from 'collections' instead "
        "of from 'collections.abc' is deprecated since "
        'Python 3.3, and in 3.9 it will stop working:'
        'DeprecationWarning:jwt.api_jwt',
        # NOTE: This looks like the line above but has a typo
        # NOTE: (a whitespace is missing):
        '-W', "ignore:Using or importing the ABCs from 'collections' instead "
        "of from 'collections.abc' is deprecated since "
        'Python 3.3,and in 3.9 it will stop working:'
        'DeprecationWarning:jwt.api_jwt',

        # NOTE: Triggered by the `octomachinery.utils.versiontools`
        # NOTE: command via `tox -e old-deps`:
        '-W', 'ignore:pkg_resources is deprecated as an API:'
        'DeprecationWarning:pkg_resources',

        # NOTE: Triggered by the `octomachinery.utils.versiontools`
        # NOTE: command via `tox -e old-deps`:
        '-W', 'ignore:The distutils package is deprecated and slated for '
        'removal in Python 3.12. Use setuptools or check PEP 632 for '
        'potential alternatives:'
        'DeprecationWarning:',

        # NOTE: Triggered by the `octomachinery.utils.versiontools`
        # NOTE: command via `tox -e old-deps`:
        '-W', "ignore:module 'sre_constants' is deprecated:"
        'DeprecationWarning:',

        # NOTE: Triggered by the `octomachinery.utils.versiontools`
        # NOTE: command via `tox -e old-deps`:
        '-W', 'ignore:_SixMetaPathImporter.exec_module() not found; '
        'falling back to load_module():ImportWarning:',

        # NOTE: Triggered by the `octomachinery.utils.versiontools`
        # NOTE: command via `tox -e old-deps`:
        '-W', 'ignore:_SixMetaPathImporter.find_spec() not found; '
        'falling back to find_module():ImportWarning:',

        # NOTE: Triggered by the `octomachinery.utils.versiontools`
        # NOTE: command via `tox -e old-deps`:
        '-W', 'ignore:VendorImporter.exec_module() not found; '
        'falling back to load_module():ImportWarning:',

        # NOTE: Triggered by the `octomachinery.utils.versiontools`
        # NOTE: command via `tox -e old-deps`:
        '-W', 'ignore:VendorImporter.find_spec() not found; '
        'falling back to find_module():ImportWarning:',

        # NOTE: Triggered by the `octomachinery.routing.routers`
        # NOTE: command via `tox -e old-deps`:
        '-W', "ignore:'cgi' is deprecated and slated for removal "
        'in Python 3.13:'
        'DeprecationWarning:gidgethub.sansio',

        '-W', 'ignore:"@coroutine" decorator is deprecated '
        'since Python 3.8, use "async def" instead:'
        'DeprecationWarning:aiohttp.helpers',
        '-c', f'import {import_path!s}',
    )

    subprocess.check_call(imp_cmd)



"""Webhook event processing helper decorators."""

from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    # pylint: disable=relative-beyond-top-level
    from ..github.models.events import GitHubEvent


__all__ = ('process_webhook_payload',)


def process_webhook_payload(wrapped_function):
    """Bypass event object keys-values as args to the handler."""
    @wraps(wrapped_function)
    def wrapper(event: GitHubEvent) -> Any:
        return wrapped_function(**event.payload)
    return wrapper
